@@ -2,7 +2,7 @@ from copy import deepcopy
 import unittest2
 from jsonobject.base import JsonObject, JsonArray
 from jsonobject import *
-from jsonobject.exceptions import DeleteNotAllowed
+from jsonobject.exceptions import DeleteNotAllowed, BadValueError
 
 
 class Features(JsonObject):
@@ -110,7 +110,7 @@ class JsonObjectTestCase(unittest2.TestCase):
 
         danny.features.hair = 'blond'
         self.assertEqual(danny.features.hair, 'blond')
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             danny.features.hair = 'green'
 
         features = {'hair': 'grey', 'eyes': 'blue'}
@@ -177,10 +177,13 @@ class JsonObjectTestCase(unittest2.TestCase):
                 eh = StringProperty(name='ay')
 
     def test_init(self):
+        from jsonobject.base import get_dynamic_properties
         self.assertEqual(JunkCD(c_property=1, d_property='yyy').to_json(),
                          JunkCD({'c': 1, 'd': 'yyy'}).to_json())
-        with self.assertRaises(AssertionError):
-            JunkCD(non_existent_property=2)
+        x = JunkCD(non_existent_property=2)
+        self.assertEqual(get_dynamic_properties(x),
+                         {'non_existent_property': 2})
+
 
         ab = JunkAB(a_property=[1, 2, 3],
                     b_property=JunkCD({'c': 1, 'd': 'string'}))
@@ -194,16 +197,16 @@ class JsonObjectTestCase(unittest2.TestCase):
 
     def test_choices(self):
 
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             Features(hair='blue')
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             Features.wrap({'hair': 'blue'})
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             f = Features()
             f.hair = 'blue'
 
     def test_required(self):
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             Person()
         Person(first_name='')
         Person(first_name='James')
@@ -211,6 +214,7 @@ class JsonObjectTestCase(unittest2.TestCase):
     def test_dynamic_properties(self):
         p = Features.wrap({'platypus': 'James'})
         p.marmot = 'Sally'
+        p._nope = 10
         self.assertEqual(p.to_json(), {
             'platypus': 'James',
             'marmot': 'Sally',
@@ -219,6 +223,7 @@ class JsonObjectTestCase(unittest2.TestCase):
         })
         self.assertEqual(p.platypus, 'James')
         self.assertEqual(p.marmot, 'Sally')
+        self.assertEqual(p._nope, 10)
 
     def test_delete_dynamic(self):
         def assertReallyThere():
@@ -263,6 +268,108 @@ class JsonObjectTestCase(unittest2.TestCase):
         self.assertEqual(json_dict, {})
         self.assertEqual(json_dict._obj, {})
 
+    def test_dynamic_container(self):
+        class Foo(JsonObject):
+            pass
+        foo = Foo(my_list=[])
+        self.assertIs(foo.my_list._obj, foo._obj['my_list'])
+        foo = Foo(my_dict={})
+        self.assertIs(foo.my_dict._obj, foo._obj['my_dict'])
+
+    def test_dynamic_dict_property(self):
+        "dates copied from couchdbkit"
+        import datetime
+
+        class Foo(JsonObject):
+            my_datetime = DateTimeProperty()
+            my_dict = DictProperty()
+        foo = Foo()
+        full_datetime = datetime.datetime(2009, 5, 10, 21, 19, 21, 127380)
+        normalized_datetime = datetime.datetime(2009, 5, 10, 21, 19, 21)
+
+        foo.my_datetime = full_datetime
+        self.assertEqual(foo.my_datetime, normalized_datetime)
+
+        foo.my_dict['test'] = {
+            'a': full_datetime
+        }
+        self.assertEqual(foo.my_dict, {
+            'test': {
+                'a': normalized_datetime
+            }
+        })
+
+
+class LazyValidationTest(unittest2.TestCase):
+
+    def _validate_raises(self, foo):
+        with self.assertRaises(BadValueError):
+            foo.validate()
+
+        with self.assertRaises(BadValueError):
+            foo.to_json()
+
+    def _validate_not_raises(self, foo):
+        foo.validate()
+        foo.to_json()
+
+    def test_string(self):
+        class Foo(JsonObject):
+            _validate_required_lazily = True
+            string = StringProperty(required=True)
+
+        foo = Foo()
+        self._validate_raises(foo)
+        foo.string = 'hi'
+        self._validate_not_raises(foo)
+
+    def test_object(self):
+        class Bar(JsonObject):
+            _validate_required_lazily = True
+            string = StringProperty(required=True)
+
+        class Foo(JsonObject):
+            _validate_required_lazily = True
+            bar = ObjectProperty(Bar)
+
+        foo = Foo()
+        self._validate_raises(foo)
+        foo.bar.string = 'hi'
+        self._validate_not_raises(foo)
+
+    def test_list(self):
+        class Bar(JsonObject):
+            _validate_required_lazily = True
+            string = StringProperty(required=True)
+
+        class Foo(JsonObject):
+            _validate_required_lazily = True
+            bars = ListProperty(Bar, required=True)
+
+        foo = Foo()
+        self._validate_raises(foo)
+        foo.bars.append(Bar())
+        self._validate_raises(foo)
+        foo.bars[0].string = 'hi'
+        self._validate_not_raises(foo)
+
+    def test_dict(self):
+        class Bar(JsonObject):
+            _validate_required_lazily = True
+            string = StringProperty(required=True)
+
+        class Foo(JsonObject):
+            _validate_required_lazily = True
+            bar_map = DictProperty(Bar, required=True)
+
+        foo = Foo()
+        self._validate_raises(foo)
+        foo.bar_map['hi'] = Bar()
+        self._validate_raises(foo)
+        foo.bar_map['hi'].string = 'hi'
+        self._validate_not_raises(foo)
+
+
 
 class PropertyTestCase(unittest2.TestCase):
     def test_date(self):
@@ -271,9 +378,9 @@ class PropertyTestCase(unittest2.TestCase):
         for string, date in [('1988-07-07', datetime.date(1988, 7, 7))]:
             self.assertEqual(p.wrap(string), date)
             self.assertEqual(p.unwrap(date), (date, string))
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             p.wrap('1234-05-90')
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             p.wrap('2000-01-01T00:00:00Z')
 
     def test_datetime(self):
@@ -282,11 +389,34 @@ class PropertyTestCase(unittest2.TestCase):
         for string, dt in [('2011-01-18T12:38:09Z', datetime.datetime(2011, 1, 18, 12, 38, 9))]:
             self.assertEqual(p.wrap(string), dt)
             self.assertEqual(p.unwrap(dt), (dt, string))
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             p.wrap('1234-05-90T00:00:00Z')
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             p.wrap('1988-07-07')
 
+    def test_time(self):
+        import datetime
+        p = TimeProperty()
+        for string, time in [('12:38:09', datetime.time(12, 38, 9))]:
+            self.assertEqual(p.wrap(string), time)
+            self.assertEqual(p.unwrap(time), (time, string))
+        with self.assertRaises(BadValueError):
+            p.wrap('25:00:00')
+        with self.assertRaises(BadValueError):
+            p.wrap('2011-01-18T12:38:09Z')
+        with self.assertRaises(BadValueError):
+            p.wrap('1988-07-07')
+
+    def test_decimal(self):
+        import decimal
+
+        class Foo(JsonObject):
+            decimal = DecimalProperty()
+
+        foo = Foo(decimal=decimal.Decimal('2.0'))
+        self.assertEqual(foo.decimal, decimal.Decimal('2.0'))
+        self.assertEqual(foo.to_json()['decimal'], '2.0')
+
     def test_dict(self):
         mapping = {'one': 1, 'two': 2}
         o = ObjectWithDictProperty(mapping=mapping)
@@ -302,7 +432,7 @@ class PropertyTestCase(unittest2.TestCase):
                 'foo': {'hair': None, 'eyes': None},
             },
         })
-        with self.assertRaises(ValueError):
+        with self.assertRaises(BadValueError):
             FeatureMap({'feature_map': {'lala': 10}})
 
         features.feature_map.update({'hoho': Features(eyes='brown')})
@@ -314,6 +444,87 @@ class PropertyTestCase(unittest2.TestCase):
             },
         })
 
+    def test_allow_dynamic(self):
+        class Foo(JsonObject):
+            _allow_dynamic_properties = False
+
+        foo = Foo()
+        with self.assertRaises(AttributeError):
+            foo.blah = 3
+        foo._blah = 5
+        self.assertEqual(foo, {})
+        self.assertEqual(foo.to_json(), {})
+        self.assertEqual(foo._blah, 5)
+
+    def test_exclude_if_none(self):
+        class Foo(JsonObject):
+            _id = StringProperty(exclude_if_none=True)
+            name = StringProperty()
+
+        foo = Foo()
+        self.assertEqual(foo.to_json(), {'name': None})
+        self.assertEqual(foo._id, None)
+        foo = Foo(_id='xxx')
+        self.assertEqual(foo, {'name': None, '_id': 'xxx'})
+        foo._id = None
+        self.assertEqual(foo.to_json(), {'name': None})
+        self.assertEqual(foo._id, None)
+
+
+class DynamicConversionTestCase(unittest2.TestCase):
+    import datetime
+
+    class Foo(JsonObject):
+        pass
+    string_date = '2012-01-01'
+    date_date = datetime.date(2012, 01, 01)
+
+    def _test_dynamic_conversion(self, foo):
+        string_date = self.string_date
+        date_date = self.date_date
+
+        self.assertEqual(foo.to_json()['my_date'], string_date)
+        self.assertEqual(foo.my_date, date_date)
+
+        self.assertEqual(foo.to_json()['my_list'], [1, 2, [string_date]])
+        self.assertEqual(foo.my_list, [1, 2, [date_date]])
+
+        self.assertEqual(foo.to_json()['my_dict'], {'a': {'b': string_date}})
+        self.assertEqual(foo.my_dict, {'a': {'b': date_date}})
+
+    def test_wrapping(self):
+        foo = self.Foo({
+            'my_date': self.string_date,
+            'my_list': [1, 2, [self.string_date]],
+            'my_dict': {'a': {'b': self.string_date}},
+        })
+        self._test_dynamic_conversion(foo)
+
+    def test_kwargs(self):
+
+        foo = self.Foo(
+            my_date=self.date_date,
+            my_list=[1, 2, [self.date_date]],
+            my_dict={'a': {'b': self.date_date}},
+        )
+        self._test_dynamic_conversion(foo)
+
+    def test_assignment(self):
+        foo = self.Foo()
+        foo.my_date = self.date_date
+        foo.my_list = [1, 2, [self.date_date]]
+        foo.my_dict = {'a': {'b': self.date_date}}
+        self._test_dynamic_conversion(foo)
+
+    def test_manipulation(self):
+        foo = self.Foo()
+        foo.my_date = self.date_date
+        foo.my_list = [1, 2, []]
+        foo.my_list[2].append(self.date_date)
+        foo.my_dict = {'a': {}}
+        foo.my_dict['a']['b'] = self.date_date
+        self._test_dynamic_conversion(foo)
+
 
 class User(JsonObject):
     username = StringProperty()
